import io
import logging
import struct
import time
import zipfile

from vtkmodules.web import (
    base64Encode,
    hashDataArray,
    getJSArrayType,
    arrayTypesMapping,
    getReferenceId,
)

from vtkmodules.vtkCommonCore import vtkTypeUInt32Array
from vtkmodules.vtkFiltersGeometry import vtkCompositeDataGeometryFilter
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter
from vtkmodules.vtkRenderingCore import vtkColorTransferFunction

logger = logging.getLogger(__name__)
# Always DEBUG level for this logger. Users can change this
logger.setLevel(logging.DEBUG)

# -----------------------------------------------------------------------------
# Array helpers
# -----------------------------------------------------------------------------

def zipCompression(name, data):
    with io.BytesIO() as in_memory:
        with zipfile.ZipFile(in_memory, mode="w") as zf:
            zf.writestr("data/%s" % name, data, zipfile.ZIP_DEFLATED)
        in_memory.seek(0)
        return in_memory.read()


def dataTableToList(dataTable):
    dataType = arrayTypesMapping[dataTable.GetDataType()]
    elementSize = struct.calcsize(dataType)
    nbValues = dataTable.GetNumberOfValues()
    nbComponents = dataTable.GetNumberOfComponents()
    nbytes = elementSize * nbValues
    if dataType != " ":
        with io.BytesIO(memoryview(dataTable)) as stream:
            data = list(struct.unpack(dataType * nbValues, stream.read(nbytes)))
        return [
            data[idx * nbComponents : (idx + 1) * nbComponents]
            for idx in range(nbValues // nbComponents)
        ]

    return None


# -----------------------------------------------------------------------------


def linspace(start, stop, num):
    delta = (stop - start) / (num - 1)
    return [start + i * delta for i in range(num)]


# -----------------------------------------------------------------------------
# Convenience class for caching data arrays, storing computed sha sums, keeping
# track of valid actors, etc...
# -----------------------------------------------------------------------------


class SynchronizationContext:
    def __init__(self):
        self.dataArrayCache = {}
        self.lastDependenciesMapping = {}
        self.ingoreLastDependencies = False

    def setIgnoreLastDependencies(self, force):
        self.ingoreLastDependencies = force

    def cacheDataArray(self, pMd5, data):
        self.dataArrayCache[pMd5] = data

    def getCachedDataArray(self, pMd5, binary=False, compression=False):
        cacheObj = self.dataArrayCache[pMd5]
        array = cacheObj["array"]
        cacheTime = cacheObj["mTime"]

        if cacheTime != array.GetMTime():
            logger.debug(" ***** ERROR: you asked for an old cache key! ***** ")

        if array.GetDataType() == 12:
            # IdType need to be converted to Uint32
            arraySize = array.GetNumberOfTuples() * array.GetNumberOfComponents()
            newArray = vtkTypeUInt32Array()
            newArray.SetNumberOfTuples(arraySize)
            for i in range(arraySize):
                newArray.SetValue(i, -1 if array.GetValue(i) < 0 else array.GetValue(i))
            pBuffer = memoryview(newArray)
        else:
            pBuffer = memoryview(array)

        if binary:
            # Convert the vtkUnsignedCharArray into a bytes object, required by
            # Autobahn websockets
            return (
                pBuffer.tobytes()
                if not compression
                else zipCompression(pMd5, pBuffer.tobytes())
            )

        return base64Encode(
            pBuffer if not compression else zipCompression(pMd5, pBuffer.tobytes())
        )

    def checkForArraysToRelease(self, timeWindow=20):
        cutOffTime = time.time() - timeWindow
        shasToDelete = []
        for sha in self.dataArrayCache:
            record = self.dataArrayCache[sha]
            array = record["array"]
            count = array.GetReferenceCount()

            if count == 1 and record["ts"] < cutOffTime:
                shasToDelete.append(sha)

        for sha in shasToDelete:
            del self.dataArrayCache[sha]

    def getLastDependencyList(self, idstr):
        lastDeps = []
        if idstr in self.lastDependenciesMapping and not self.ingoreLastDependencies:
            lastDeps = self.lastDependenciesMapping[idstr]
        return lastDeps

    def setNewDependencyList(self, idstr, depList):
        self.lastDependenciesMapping[idstr] = depList

    def buildDependencyCallList(self, idstr, newList, addMethod, removeMethod):
        oldList = self.getLastDependencyList(idstr)

        calls = []
        calls += [[addMethod, [wrapId(x)]] for x in newList if x not in oldList]
        calls += [[removeMethod, [wrapId(x)]] for x in oldList if x not in newList]

        self.setNewDependencyList(idstr, newList)
        return calls


# -----------------------------------------------------------------------------
# Global variables
# -----------------------------------------------------------------------------

SERIALIZERS = {}
JS_CLASS_MAPPING = {}
context = None

# -----------------------------------------------------------------------------
# Global API
# -----------------------------------------------------------------------------


def registerInstanceSerializer(name, method):
    global SERIALIZERS
    SERIALIZERS[name] = method

def registerJSClass(vtk_class, js_class):
    global JS_CLASS_MAPPING
    JS_CLASS_MAPPING[vtk_class] = js_class

def class_name(vtk_obj):
    vtk_class = vtk_obj.GetClassName()
    if vtk_class in JS_CLASS_MAPPING:
        return JS_CLASS_MAPPING[vtk_class]

    return vtk_class


# -----------------------------------------------------------------------------


def serializeInstance(parent, instance, instanceId, context, depth):
    instanceType = class_name(instance)
    serializer = SERIALIZERS[instanceType] if instanceType in SERIALIZERS else None

    if serializer:
        return serializer(parent, instance, instanceId, context, depth)

    logger.error(f"!!!No serializer for {instanceType} with id {instanceId}")

    return None


# -----------------------------------------------------------------------------


def initializeSerializers():
    # Actors/viewProps
    registerInstanceSerializer("vtkActor", genericActorSerializer)
    registerInstanceSerializer("vtkOpenGLActor", genericActorSerializer)
    registerInstanceSerializer("vtkPVLODActor", genericActorSerializer)

    # Volume/viewProps
    registerInstanceSerializer("vtkVolume", genericVolumeSerializer)

    # Mappers
    registerInstanceSerializer("vtkMapper", genericMapperSerializer)
    registerInstanceSerializer("vtkDataSetMapper", genericMapperSerializer)
    registerInstanceSerializer("vtkPolyDataMapper", genericMapperSerializer)
    registerInstanceSerializer("vtkImageDataMapper", genericMapperSerializer)
    registerInstanceSerializer("vtkOpenGLPolyDataMapper", genericMapperSerializer)
    registerInstanceSerializer("vtkCompositePolyDataMapper2", genericMapperSerializer)
    registerJSClass("vtkPolyDataMapper", "vtkMapper")
    registerJSClass("vtkDataSetMapper", "vtkMapper")
    registerJSClass("vtkOpenGLPolyDataMapper", "vtkMapper")
    registerJSClass("vtkCompositePolyDataMapper2", "vtkMapper")

    registerInstanceSerializer("vtkVolumeMapper", genericVolumeMapperSerializer)
    registerInstanceSerializer("vtkFixedPointVolumeRayCastMapper", genericVolumeMapperSerializer)
    registerJSClass("vtkFixedPointVolumeRayCastMapper", "vtkVolumeMapper")

    # LookupTables/TransferFunctions
    registerInstanceSerializer("vtkLookupTable", lookupTableSerializer2)
    registerInstanceSerializer(
        "vtkPVDiscretizableColorTransferFunction", discretizableColorTransferFunctionSerializer
    )
    registerInstanceSerializer(
        "vtkColorTransferFunction", colorTransferFunctionSerializer
    )
    registerInstanceSerializer("vtkPiecewiseFunction", pwfSerializer)

    # Textures
    registerInstanceSerializer("vtkTexture", textureSerializer)
    registerInstanceSerializer("vtkOpenGLTexture", textureSerializer)

    # Property
    registerInstanceSerializer("vtkProperty", propertySerializer)
    registerInstanceSerializer("vtkOpenGLProperty", propertySerializer)

    # VolumeProperty
    registerInstanceSerializer("vtkVolumeProperty", volumePropertySerializer)

    # Datasets
    registerInstanceSerializer("vtkPolyData", polydataSerializer)
    registerInstanceSerializer("vtkImageData", imagedataSerializer)
    registerInstanceSerializer("vtkUnstructuredGrid", mergeToPolydataSerializer)
    registerInstanceSerializer("vtkMultiBlockDataSet", mergeToPolydataSerializer)
    registerInstanceSerializer("vtkStructuredPoints", imagedataSerializer)
    registerJSClass("vtkStructuredPoints", "vtkImageData")


    # RenderWindows
    registerInstanceSerializer("vtkRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkCocoaRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkXOpenGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkWin32OpenGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkEGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkOpenVRRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkOpenXRRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkGenericOpenGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkOSOpenGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkOpenGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkIOSRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkExternalOpenGLRenderWindow", renderWindowSerializer)
    registerInstanceSerializer("vtkOffscreenOpenGLRenderWindow", renderWindowSerializer)

    # Renderers
    registerInstanceSerializer("vtkRenderer", rendererSerializer)
    registerInstanceSerializer("vtkOpenGLRenderer", rendererSerializer)

    # Cameras
    registerInstanceSerializer("vtkCamera", cameraSerializer)
    registerInstanceSerializer("vtkOpenGLCamera", cameraSerializer)

    # Lights
    registerInstanceSerializer("vtkLight", lightSerializer)
    registerInstanceSerializer("vtkPVLight", lightSerializer)
    registerInstanceSerializer("vtkOpenGLLight", lightSerializer)

    # Annotations (ScalarBar/CubeAxes
    registerInstanceSerializer("vtkCubeAxesActor", cubeAxesSerializer)
    registerInstanceSerializer("vtkScalarBarActor", scalarBarActorSerializer)


# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------


def pad(depth):
    padding = ""
    for _ in range(depth):
        padding += "  "
    return padding


# -----------------------------------------------------------------------------


def wrapId(idStr):
    return "instance:${%s}" % idStr


# -----------------------------------------------------------------------------

dataArrayShaMapping = {}


def digest(array):
    objId = getReferenceId(array)

    record = None
    if objId in dataArrayShaMapping:
        record = dataArrayShaMapping[objId]

    if record and record["mtime"] == array.GetMTime():
        return record["sha"]

    record = {"sha": hashDataArray(array), "mtime": array.GetMTime()}

    dataArrayShaMapping[objId] = record
    return record["sha"]


# -----------------------------------------------------------------------------


def getRangeInfo(array, component):
    r = array.GetRange(component)
    compRange = {}
    compRange["min"] = r[0]
    compRange["max"] = r[1]
    compRange["component"] = array.GetComponentName(component)
    return compRange


# -----------------------------------------------------------------------------


def getArrayDescription(array, context):
    if not array:
        return None

    pMd5 = digest(array)
    context.cacheDataArray(
        pMd5, {"array": array, "mTime": array.GetMTime(), "ts": time.time()}
    )

    root = {}
    root["hash"] = pMd5
    root["vtkClass"] = "vtkDataArray"
    root["name"] = array.GetName()
    root["dataType"] = getJSArrayType(array)
    root["numberOfComponents"] = array.GetNumberOfComponents()
    root["size"] = array.GetNumberOfComponents() * array.GetNumberOfTuples()
    root["ranges"] = []
    if root["numberOfComponents"] > 1:
        for i in range(root["numberOfComponents"]):
            root["ranges"].append(getRangeInfo(array, i))
        root["ranges"].append(getRangeInfo(array, -1))
    else:
        root["ranges"].append(getRangeInfo(array, 0))

    return root


# -----------------------------------------------------------------------------


def extractRequiredFields(
    extractedFields, parent, dataset, context, requestedFields=["Normals", "TCoords"]
):
    arrays_to_export = set()
    export_all = "*" in requestedFields
    # Identify arrays to export
    if not export_all:
        # FIXME should evolve and support funky mapper which leverage many arrays
        if parent and parent.IsA("vtkMapper"):
            mapper = parent
            scalarVisibility = mapper.GetScalarVisibility()
            arrayAccessMode = mapper.GetArrayAccessMode()
            colorArrayName = (
                mapper.GetArrayName() if arrayAccessMode == 1 else mapper.GetArrayId()
            )
            # colorMode = mapper.GetColorMode()
            scalarMode = mapper.GetScalarMode()
            if scalarVisibility and scalarMode in (1, 3):
                array_to_export = dataset.GetPointData().GetArray(colorArrayName)
                if array_to_export is None:
                    array_to_export = dataset.GetPointData().GetScalars()
                arrays_to_export.add(array_to_export)
            if scalarVisibility and scalarMode in (2, 4):
                array_to_export = dataset.GetCellData().GetArray(colorArrayName)
                if array_to_export is None:
                    array_to_export = dataset.GetCellData().GetScalars()
                arrays_to_export.add(array_to_export)
            if scalarVisibility and scalarMode == 0:
                array_to_export = dataset.GetPointData().GetScalars()
                if array_to_export is None:
                    array_to_export = dataset.GetCellData().GetScalars()
                arrays_to_export.add(array_to_export)

        if parent and parent.IsA("vtkTexture") and dataset.GetPointData().GetScalars():
            arrays_to_export.add(dataset.GetPointData().GetScalars())

        arrays_to_export.update(
            [
                getattr(dataset.GetPointData(), "Get" + requestedField, lambda: None)()
                for requestedField in requestedFields
            ]
        )

    # Browse all arrays
    for location, field_data in [
        ("pointData", dataset.GetPointData()),
        ("cellData", dataset.GetCellData()),
    ]:
        for array_index in range(field_data.GetNumberOfArrays()):
            array = field_data.GetArray(array_index)
            if export_all or array in arrays_to_export:
                arrayMeta = getArrayDescription(array, context)
                if arrayMeta:
                    arrayMeta["location"] = location
                    attribute = field_data.IsArrayAnAttribute(array_index)
                    arrayMeta["registration"] = (
                        "set" + field_data.GetAttributeTypeAsString(attribute)
                        if attribute >= 0
                        else "addArray"
                    )
                    extractedFields.append(arrayMeta)

# -----------------------------------------------------------------------------
# Concrete instance serializers
# -----------------------------------------------------------------------------


def genericActorSerializer(parent, actor, actorId, context, depth):
    # This kind of actor has two "children" of interest, a property and a
    # mapper
    actorVisibility = actor.GetVisibility()
    mapperInstance = None
    propertyInstance = None
    calls = []
    dependencies = []

    if actorVisibility:
        mapper = None
        if not hasattr(actor, "GetMapper"):
            logger.debug("This actor does not have a GetMapper method")
        else:
            mapper = actor.GetMapper()

        if mapper:
            mapperId = getReferenceId(mapper)
            mapperInstance = serializeInstance(
                actor, mapper, mapperId, context, depth + 1
            )
            if mapperInstance:
                dependencies.append(mapperInstance)
                calls.append(["setMapper", [wrapId(mapperId)]])

        prop = None
        if hasattr(actor, "GetProperty"):
            prop = actor.GetProperty()
        else:
            logger.debug("This actor does not have a GetProperty method")

        if prop:
            propId = getReferenceId(prop)
            propertyInstance = serializeInstance(
                actor, prop, propId, context, depth + 1
            )
            if propertyInstance:
                dependencies.append(propertyInstance)
                calls.append(["setProperty", [wrapId(propId)]])

        # Handle texture if any
        texture = None
        if hasattr(actor, "GetTexture"):
            texture = actor.GetTexture()
        else:
            logger.debug("This actor does not have a GetTexture method")

        if texture:
            textureId = getReferenceId(texture)
            textureInstance = serializeInstance(
                actor, texture, textureId, context, depth + 1
            )
            if textureInstance:
                dependencies.append(textureInstance)
                calls.append(["addTexture", [wrapId(textureId)]])

    if actorVisibility == 0 or (mapperInstance and propertyInstance):
        return {
            "parent": getReferenceId(parent),
            "id": actorId,
            "type": class_name(actor),
            "properties": {
                # vtkProp
                "visibility": actorVisibility,
                "pickable": actor.GetPickable(),
                "dragable": actor.GetDragable(),
                "useBounds": actor.GetUseBounds(),
                # vtkProp3D
                "origin": actor.GetOrigin(),
                "position": actor.GetPosition(),
                "scale": actor.GetScale(),
                # vtkActor
                "forceOpaque": actor.GetForceOpaque(),
                "forceTranslucent": actor.GetForceTranslucent(),
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None


# -----------------------------------------------------------------------------


def genericVolumeSerializer(parent, actor, actorId, context, depth):
    # This kind of actor has two "children" of interest, a property and a
    # mapper
    actorVisibility = actor.GetVisibility()
    mapperInstance = None
    propertyInstance = None
    calls = []
    dependencies = []

    if actorVisibility:
        mapper = None
        if not hasattr(actor, "GetMapper"):
            logger.debug("This actor does not have a GetMapper method")
        else:
            mapper = actor.GetMapper()

        if mapper:
            mapperId = getReferenceId(mapper)
            mapperInstance = serializeInstance(
                actor, mapper, mapperId, context, depth + 1
            )
            if mapperInstance:
                dependencies.append(mapperInstance)
                calls.append(["setMapper", [wrapId(mapperId)]])

        prop = None
        if hasattr(actor, "GetProperty"):
            prop = actor.GetProperty()
        else:
            logger.debug("This actor does not have a GetProperty method")

        if prop:
            propId = getReferenceId(prop)
            propertyInstance = serializeInstance(
                actor, prop, propId, context, depth + 1
            )
            if propertyInstance:
                dependencies.append(propertyInstance)
                calls.append(["setProperty", [wrapId(propId)]])

    if actorVisibility == 0 or (mapperInstance and propertyInstance):
        return {
            "parent": getReferenceId(parent),
            "id": actorId,
            "type": class_name(actor),
            "properties": {
                # vtkProp
                "visibility": actorVisibility,
                "pickable": actor.GetPickable(),
                "dragable": actor.GetDragable(),
                "useBounds": actor.GetUseBounds(),
                # vtkProp3D
                "origin": actor.GetOrigin(),
                "position": actor.GetPosition(),
                "scale": actor.GetScale(),
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None

# -----------------------------------------------------------------------------


def textureSerializer(parent, texture, textureId, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    dataObject = None
    dataObjectInstance = None
    calls = []
    dependencies = []

    if hasattr(texture, "GetInput"):
        dataObject = texture.GetInput()
    else:
        logger.debug("This texture does not have GetInput method")

    if dataObject:
        dataObjectId = "%s-texture" % textureId
        dataObjectInstance = serializeInstance(
            texture, dataObject, dataObjectId, context, depth + 1
        )
        if dataObjectInstance:
            dependencies.append(dataObjectInstance)
            calls.append(["setInputData", [wrapId(dataObjectId)]])

    if dataObjectInstance:
        return {
            "parent": getReferenceId(parent),
            "id": textureId,
            "type": "vtkTexture",
            "properties": {
                "interpolate": texture.GetInterpolate(),
                "repeat": texture.GetRepeat(),
                "edgeClamp": texture.GetEdgeClamp(),
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None


# -----------------------------------------------------------------------------


def genericMapperSerializer(parent, mapper, mapperId, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    dataObject = None
    dataObjectInstance = None
    lookupTableInstance = None
    calls = []
    dependencies = []

    if hasattr(mapper, "GetInputDataObject"):
        mapper.GetInputAlgorithm().Update()
        dataObject = mapper.GetInputDataObject(0, 0)
    else:
        logger.debug("This mapper does not have GetInputDataObject method")

    if dataObject:
        if dataObject.IsA("vtkDataSet"):
            alg = vtkDataSetSurfaceFilter()
            alg.SetInputData(dataObject)
            alg.Update()
            dataObject = alg.GetOutput()

        dataObjectId = "%s-dataset" % mapperId
        dataObjectInstance = serializeInstance(
            mapper, dataObject, dataObjectId, context, depth + 1
        )

        if dataObjectInstance:
            dependencies.append(dataObjectInstance)
            calls.append(["setInputData", [wrapId(dataObjectId)]])

    lookupTable = None

    if hasattr(mapper, "GetLookupTable"):
        lookupTable = mapper.GetLookupTable()
    else:
        logger.debug("This mapper does not have GetLookupTable method")

    if lookupTable:
        lookupTableId = getReferenceId(lookupTable)
        lookupTableInstance = serializeInstance(
            mapper, lookupTable, lookupTableId, context, depth + 1
        )
        if lookupTableInstance:
            dependencies.append(lookupTableInstance)
            calls.append(
                ["setLookupTable", [wrapId(lookupTableId)]]
            )

    if dataObjectInstance:
        colorArrayName = (
            mapper.GetArrayName()
            if mapper.GetArrayAccessMode() == 1
            else mapper.GetArrayId()
        )
        return {
            "parent": getReferenceId(parent),
            "id": mapperId,
            "type": class_name(mapper),
            "properties": {
                "resolveCoincidentTopology": mapper.GetResolveCoincidentTopology(),
                "renderTime": mapper.GetRenderTime(),
                "arrayAccessMode": mapper.GetArrayAccessMode(),
                "scalarRange": mapper.GetScalarRange(),
                "useLookupTableScalarRange": 1
                if mapper.GetUseLookupTableScalarRange()
                else 0,
                "scalarVisibility": mapper.GetScalarVisibility(),
                "colorByArrayName": colorArrayName,
                "colorMode": mapper.GetColorMode(),
                "scalarMode": mapper.GetScalarMode(),
                "interpolateScalarsBeforeMapping": 1
                if mapper.GetInterpolateScalarsBeforeMapping()
                else 0,
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None


# -----------------------------------------------------------------------------


def genericVolumeMapperSerializer(parent, mapper, mapperId, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    dataObject = None
    dataObjectInstance = None
    lookupTableInstance = None
    calls = []
    dependencies = []

    if hasattr(mapper, "GetInputDataObject"):
        mapper.GetInputAlgorithm().Update()
        dataObject = mapper.GetInputDataObject(0, 0)
    else:
        logger.debug("This mapper does not have GetInputDataObject method")

    if dataObject:
        dataObjectId = "%s-dataset" % mapperId
        dataObjectInstance = serializeInstance(
            mapper, dataObject, dataObjectId, context, depth + 1
        )

        if dataObjectInstance:
            dependencies.append(dataObjectInstance)
            calls.append(["setInputData", [wrapId(dataObjectId)]])

    if dataObjectInstance:
        return {
            "parent": getReferenceId(parent),
            "id": mapperId,
            "type": class_name(mapper),
            "properties": {
                # VolumeMapper
                "sampleDistance": mapper.GetSampleDistance(),
                "imageSampleDistance": mapper.GetImageSampleDistance(),
                # "maximumSamplesPerRay": mapper.GetMaximumSamplesPerRay(),
                "autoAdjustSampleDistances": mapper.GetAutoAdjustSampleDistances(),
                "blendMode": mapper.GetBlendMode(),
                # "ipScalarRange": mapper.GetIpScalarRange(),
                # "filterMode": mapper.GetFilterMode(),
                # "preferSizeOverAccuracy": mapper.Get(),
            },
            "calls": calls,
            "dependencies": dependencies,
        }

    return None

# -----------------------------------------------------------------------------


def lookupTableSerializer(parent, lookupTable, lookupTableId, context, depth):
    # No children in this case, so no additions to bindings and return empty list
    # But we do need to add instance

    lookupTableRange = lookupTable.GetRange()

    lookupTableHueRange = [0.5, 0]
    if hasattr(lookupTable, "GetHueRange"):
        try:
            lookupTable.GetHueRange(lookupTableHueRange)
        except Exception as inst:
            pass

    lutSatRange = lookupTable.GetSaturationRange()
    lutAlphaRange = lookupTable.GetAlphaRange()

    return {
        "parent": getReferenceId(parent),
        "id": lookupTableId,
        "type": class_name(lookupTable),
        "properties": {
            "numberOfColors": lookupTable.GetNumberOfColors(),
            "valueRange": lookupTableRange,
            "hueRange": lookupTableHueRange,
            # 'alphaRange': lutAlphaRange,  # Causes weird rendering artifacts on client
            "saturationRange": lutSatRange,
            "nanColor": lookupTable.GetNanColor(),
            "belowRangeColor": lookupTable.GetBelowRangeColor(),
            "aboveRangeColor": lookupTable.GetAboveRangeColor(),
            "useAboveRangeColor": True
            if lookupTable.GetUseAboveRangeColor()
            else False,
            "useBelowRangeColor": True
            if lookupTable.GetUseBelowRangeColor()
            else False,
            "alpha": lookupTable.GetAlpha(),
            "vectorSize": lookupTable.GetVectorSize(),
            "vectorComponent": lookupTable.GetVectorComponent(),
            "vectorMode": lookupTable.GetVectorMode(),
            "indexedLookup": lookupTable.GetIndexedLookup(),
        },
    }


# -----------------------------------------------------------------------------


def lookupTableToColorTransferFunction(lookupTable):
    dataTable = lookupTable.GetTable()
    table = dataTableToList(dataTable)
    if table:
        ctf = vtkColorTransferFunction()
        tableRange = lookupTable.GetTableRange()
        points = linspace(*tableRange, num=len(table))
        for x, rgba in zip(points, table):
            ctf.AddRGBPoint(x, *[x / 255 for x in rgba[:3]])

        return ctf

    return None


def lookupTableSerializer2(parent, lookupTable, lookupTableId, context, depth):
    ctf = lookupTableToColorTransferFunction(lookupTable)
    if ctf:
        return colorTransferFunctionSerializer(
            parent, ctf, lookupTableId, context, depth
        )

    return None


# -----------------------------------------------------------------------------


def propertySerializer(parent, propObj, propObjId, context, depth):
    representation = (
        propObj.GetRepresentation() if hasattr(propObj, "GetRepresentation") else 2
    )
    colorToUse = (
        propObj.GetDiffuseColor() if hasattr(propObj, "GetDiffuseColor") else [1, 1, 1]
    )
    if representation == 1 and hasattr(propObj, "GetColor"):
        colorToUse = propObj.GetColor()

    return {
        "parent": getReferenceId(parent),
        "id": propObjId,
        "type": class_name(propObj),
        "properties": {
            "representation": representation,
            "diffuseColor": colorToUse,
            "color": propObj.GetColor(),
            "ambientColor": propObj.GetAmbientColor(),
            "specularColor": propObj.GetSpecularColor(),
            "edgeColor": propObj.GetEdgeColor(),
            "ambient": propObj.GetAmbient(),
            "diffuse": propObj.GetDiffuse(),
            "specular": propObj.GetSpecular(),
            "specularPower": propObj.GetSpecularPower(),
            "opacity": propObj.GetOpacity(),
            "interpolation": propObj.GetInterpolation(),
            "edgeVisibility": 1 if propObj.GetEdgeVisibility() else 0,
            "backfaceCulling": 1 if propObj.GetBackfaceCulling() else 0,
            "frontfaceCulling": 1 if propObj.GetFrontfaceCulling() else 0,
            "pointSize": propObj.GetPointSize(),
            "lineWidth": propObj.GetLineWidth(),
            "lighting": 1 if propObj.GetLighting() else 0,
        },
    }

def volumePropertySerializer(parent, propObj, propObjId, context, depth):
    calls = []
    dependencies = []

    # Color handling
    lut = propObj.GetRGBTransferFunction()
    if lut:
        lookupTableId = getReferenceId(lut)
        lookupTableInstance = serializeInstance(
            propObj, lut, lookupTableId, context, depth + 1
        )

        if lookupTableInstance:
            dependencies.append(lookupTableInstance)
            calls.append(["setRGBTransferFunction", [0, wrapId(lookupTableId)]])

    # Piecewise handling
    pwf = propObj.GetScalarOpacity()
    if pwf:
        pwfId = getReferenceId(pwf)
        pwfInstance = serializeInstance(
            propObj, pwf, pwfId, context, depth + 1
        )

        if pwfInstance:
            dependencies.append(pwfInstance)
            calls.append(["setScalarOpacity", [0, wrapId(pwfId)]])

    return {
        "parent": getReferenceId(parent),
        "id": propObjId,
        "type": class_name(propObj),
        "properties": {
            "independentComponents": propObj.GetIndependentComponents(),
            "interpolationType": propObj.GetInterpolationType(),
            "shade": propObj.GetShade(),
            "ambient": propObj.GetAmbient(),
            "diffuse": propObj.GetDiffuse(),
            "specular": propObj.GetSpecular(),
            "specularPower": propObj.GetSpecularPower(),
            # "useLabelOutline": propObj.GetUseLabelOutline(),
            # "labelOutlineThickness": propObj.GetLabelOutlineThickness(),
        },
        "calls": calls,
        "dependencies": dependencies,
    }

# -----------------------------------------------------------------------------


def imagedataSerializer(parent, dataset, datasetId, context, depth, requested_fields = ["Normals", "TCoords"]):
    if hasattr(dataset, "GetDirectionMatrix"):
        direction = [dataset.GetDirectionMatrix().GetElement(0, i) for i in range(9)]
    else:
        direction = [1, 0, 0, 0, 1, 0, 0, 0, 1]

    # Extract dataset fields
    fields = []
    extractRequiredFields(fields, parent, dataset, context, "*")

    return {
        "parent": getReferenceId(parent),
        "id": datasetId,
        "type": class_name(dataset),
        "properties": {
            "spacing": dataset.GetSpacing(),
            "origin": dataset.GetOrigin(),
            "dimensions": dataset.GetDimensions(),
            "direction": direction,
            "fields": fields,
        },
    }


# -----------------------------------------------------------------------------


def polydataSerializer(parent, dataset, datasetId, context, depth, requested_fields = ["Normals", "TCoords"]):
    if dataset and dataset.GetPoints():
        properties = {}

        # Points
        points = getArrayDescription(dataset.GetPoints().GetData(), context)
        points["vtkClass"] = "vtkPoints"
        properties["points"] = points

        # Verts
        if dataset.GetVerts() and dataset.GetVerts().GetData().GetNumberOfTuples() > 0:
            _verts = getArrayDescription(dataset.GetVerts().GetData(), context)
            properties["verts"] = _verts
            properties["verts"]["vtkClass"] = "vtkCellArray"

        # Lines
        if dataset.GetLines() and dataset.GetLines().GetData().GetNumberOfTuples() > 0:
            _lines = getArrayDescription(dataset.GetLines().GetData(), context)
            properties["lines"] = _lines
            properties["lines"]["vtkClass"] = "vtkCellArray"

        # Polys
        if dataset.GetPolys() and dataset.GetPolys().GetData().GetNumberOfTuples() > 0:
            _polys = getArrayDescription(dataset.GetPolys().GetData(), context)
            properties["polys"] = _polys
            properties["polys"]["vtkClass"] = "vtkCellArray"

        # Strips
        if (
            dataset.GetStrips()
            and dataset.GetStrips().GetData().GetNumberOfTuples() > 0
        ):
            _strips = getArrayDescription(dataset.GetStrips().GetData(), context)
            properties["strips"] = _strips
            properties["strips"]["vtkClass"] = "vtkCellArray"

        # Fields
        properties["fields"] = []
        extractRequiredFields(properties["fields"], parent, dataset, context, requested_fields)

        return {
            "parent": getReferenceId(parent),
            "id": datasetId,
            "type": class_name(dataset),
            "properties": properties,
        }

    logger.debug("This dataset has no points!")
    return None


# -----------------------------------------------------------------------------


def mergeToPolydataSerializer(parent, dataObject, dataObjectId, context, depth, requested_fields=["Normals", "TCoords"]):
    dataset = None

    if dataObject.IsA("vtkCompositeDataSet"):
        gf = vtkCompositeDataGeometryFilter()
        gf.SetInputData(dataObject)
        gf.Update()
        dataset = gf.GetOutput()
    elif dataObject.IsA("vtkUnstructuredGrid"):
        gf = vtkDataSetSurfaceFilter()
        gf.SetInputData(dataObject)
        gf.Update()
        dataset = gf.GetOutput()
    else:
        dataset = mapper.GetInput()

    return polydataSerializer(parent, dataset, dataObjectId, context, depth, requested_fields)


# -----------------------------------------------------------------------------


def colorTransferFunctionSerializer(parent, instance, objId, context, depth):
    nodes = []

    for i in range(instance.GetSize()):
        # x, r, g, b, midpoint, sharpness
        node = [0, 0, 0, 0, 0, 0]
        instance.GetNodeValue(i, node)
        nodes.append(node)

    return {
        "parent": getReferenceId(parent),
        "id": objId,
        "type": class_name(instance),
        "properties": {
            "clamping": 1 if instance.GetClamping() else 0,
            "colorSpace": instance.GetColorSpace(),
            "hSVWrap": 1 if instance.GetHSVWrap() else 0,
            # 'nanColor': instance.GetNanColor(),                  # Breaks client
            # 'belowRangeColor': instance.GetBelowRangeColor(),    # Breaks client
            # 'aboveRangeColor': instance.GetAboveRangeColor(),    # Breaks client
            # 'useAboveRangeColor': 1 if instance.GetUseAboveRangeColor() else 0,
            # 'useBelowRangeColor': 1 if instance.GetUseBelowRangeColor() else 0,
            "allowDuplicateScalars": 1 if instance.GetAllowDuplicateScalars() else 0,
            "alpha": instance.GetAlpha(),
            "vectorComponent": instance.GetVectorComponent(),
            "vectorSize": instance.GetVectorSize(),
            "vectorMode": instance.GetVectorMode(),
            "indexedLookup": instance.GetIndexedLookup(),
            "nodes": nodes,
        },
    }

def discretizableColorTransferFunctionSerializer(parent, instance, objId, context, depth):
    ctf = colorTransferFunctionSerializer(parent, instance, objId, context, depth)
    ctf["properties"]["discretize"] = instance.GetDiscretize()
    ctf["properties"]["numberOfValues"] = instance.GetNumberOfValues()
    return ctf

# -----------------------------------------------------------------------------

def pwfSerializer(parent, instance, objId, context, depth):
    nodes = []

    for i in range(instance.GetSize()):
        # x, y, midpoint, sharpness
        node = [0, 0, 0, 0]
        instance.GetNodeValue(i, node)
        nodes.append(node)

    return {
        "parent": getReferenceId(parent),
        "id": objId,
        "type": class_name(instance),
        "properties": {
            "range": list(instance.GetRange()),
            "clamping": instance.GetClamping(),
            "allowDuplicateScalars": instance.GetAllowDuplicateScalars(),
            "nodes": nodes,
        },
    }

# -----------------------------------------------------------------------------

def cubeAxesSerializer(parent, actor, actorId, context, depth):
    """
    Possible add-on properties for vtk.js:
        gridLines: True,
        axisLabels: None,
        axisTitlePixelOffset: 35.0,
        axisTextStyle: {
            fontColor: 'white',
            fontStyle: 'normal',
            fontSize: 18,
            fontFamily: 'serif',
        },
        tickLabelPixelOffset: 12.0,
        tickTextStyle: {
            fontColor: 'white',
            fontStyle: 'normal',
            fontSize: 14,
            fontFamily: 'serif',
        },
    """
    axisLabels = ["", "", ""]
    if actor.GetXAxisLabelVisibility():
        axisLabels[0] = actor.GetXTitle()
    if actor.GetYAxisLabelVisibility():
        axisLabels[1] = actor.GetYTitle()
    if actor.GetZAxisLabelVisibility():
        axisLabels[2] = actor.GetZTitle()

    return {
        "parent": getReferenceId(parent),
        "id": actorId,
        "type": "vtkCubeAxesActor",
        "properties": {
            # vtkProp
            "visibility": actor.GetVisibility(),
            "pickable": actor.GetPickable(),
            "dragable": actor.GetDragable(),
            "useBounds": actor.GetUseBounds(),
            # vtkProp3D
            "origin": actor.GetOrigin(),
            "position": actor.GetPosition(),
            "scale": actor.GetScale(),
            # vtkActor
            "forceOpaque": actor.GetForceOpaque(),
            "forceTranslucent": actor.GetForceTranslucent(),
            # vtkCubeAxesActor
            "dataBounds": actor.GetBounds(),
            "faceVisibilityAngle": 8,
            "gridLines": True,
            "axisLabels": axisLabels,
            "axisTitlePixelOffset": 35.0,
            "axisTextStyle": {
                "fontColor": "white",
                "fontStyle": "normal",
                "fontSize": 18,
                "fontFamily": "serif",
            },
            "tickLabelPixelOffset": 12.0,
            "tickTextStyle": {
                "fontColor": "white",
                "fontStyle": "normal",
                "fontSize": 14,
                "fontFamily": "serif",
            },
        },
        "calls": [["setCamera", [wrapId(getReferenceId(actor.GetCamera()))]]],
        "dependencies": [],
    }

# -----------------------------------------------------------------------------

def scalarBarActorSerializer(parent, actor, actorId, context, depth):
    dependencies = []
    calls = []
    lut = actor.GetLookupTable()
    if not lut:
        return None

    lutId = getReferenceId(lut)
    lutInstance = serializeInstance(actor, lut, lutId, context, depth + 1)
    if not lutInstance:
        return None

    dependencies.append(lutInstance)
    calls.append(["setScalarsToColors", [wrapId(lutId)]])

    prop = None
    if hasattr(actor, "GetProperty"):
        prop = actor.GetProperty()
    else:
        logger.debug("This scalarBarActor does not have a GetProperty method")

        if prop:
            propId = getReferenceId(prop)
            propertyInstance = serializeInstance(
                actor, prop, propId, context, depth + 1
            )
            if propertyInstance:
                dependencies.append(propertyInstance)
                calls.append(["setProperty", [wrapId(propId)]])

    axisLabel = actor.GetTitle()
    width = actor.GetWidth()
    height = actor.GetHeight()

    return {
        "parent": getReferenceId(parent),
        "id": actorId,
        "type": "vtkScalarBarActor",
        "properties": {
            # vtkProp
            "visibility": actor.GetVisibility(),
            "pickable": actor.GetPickable(),
            "dragable": actor.GetDragable(),
            "useBounds": actor.GetUseBounds(),
            # vtkActor2D
            # "position": actor.GetPosition(),
            # "position2": actor.GetPosition2(),
            # "width": actor.GetWidth(),
            # "height": actor.GetHeight(),
            # vtkScalarBarActor
            "automated": True,
            "axisLabel": axisLabel,
            # 'barPosition': [0, 0],
            # 'barSize': [0, 0],
            "boxPosition": [0.88, -0.92],
            "boxSize": [width, height],
            "axisTitlePixelOffset": 36.0,
            "axisTextStyle": {
                "fontColor": actor.GetTitleTextProperty().GetColor(),
                "fontStyle": "normal",
                "fontSize": 18,
                "fontFamily": "serif",
            },
            "tickLabelPixelOffset": 14.0,
            "tickTextStyle": {
                "fontColor": actor.GetTitleTextProperty().GetColor(),
                "fontStyle": "normal",
                "fontSize": 14,
                "fontFamily": "serif",
            },
            "drawNanAnnotation": actor.GetDrawNanAnnotation(),
            "drawBelowRangeSwatch": actor.GetDrawBelowRangeSwatch(),
            "drawAboveRangeSwatch": actor.GetDrawAboveRangeSwatch(),
        },
        "calls": calls,
        "dependencies": dependencies,
    }

# -----------------------------------------------------------------------------


def rendererSerializer(parent, instance, objId, context, depth):
    dependencies = []
    viewPropIds = []
    lightsIds = []
    calls = []

    # Camera
    camera = instance.GetActiveCamera()
    cameraId = getReferenceId(camera)
    cameraInstance = serializeInstance(instance, camera, cameraId, context, depth + 1)
    if cameraInstance:
        dependencies.append(cameraInstance)
        calls.append(["setActiveCamera", [wrapId(cameraId)]])

    # View prop as representation containers
    viewPropCollection = instance.GetViewProps()
    for rpIdx in range(viewPropCollection.GetNumberOfItems()):
        viewProp = viewPropCollection.GetItemAsObject(rpIdx)
        viewPropId = getReferenceId(viewProp)

        viewPropInstance = serializeInstance(
            instance, viewProp, viewPropId, context, depth + 1
        )
        if viewPropInstance:
            dependencies.append(viewPropInstance)
            viewPropIds.append(viewPropId)

    calls += context.buildDependencyCallList(
        "%s-props" % objId, viewPropIds, "addViewProp", "removeViewProp"
    )

    # Lights
    lightCollection = instance.GetLights()
    for lightIdx in range(lightCollection.GetNumberOfItems()):
        light = lightCollection.GetItemAsObject(lightIdx)
        lightId = getReferenceId(light)

        lightInstance = serializeInstance(instance, light, lightId, context, depth + 1)
        if lightInstance:
            dependencies.append(lightInstance)
            lightsIds.append(lightId)

    calls += context.buildDependencyCallList(
        "%s-lights" % objId, lightsIds, "addLight", "removeLight"
    )

    if len(dependencies) > 1:
        return {
            "parent": getReferenceId(parent),
            "id": objId,
            "type": class_name(instance),
            "properties": {
                "background": instance.GetBackground(),
                "background2": instance.GetBackground2(),
                "viewport": instance.GetViewport(),
                # These commented properties do not yet have real setters in vtk.js
                # 'gradientBackground': instance.GetGradientBackground(),
                # 'aspect': instance.GetAspect(),
                # 'pixelAspect': instance.GetPixelAspect(),
                # 'ambient': instance.GetAmbient(),
                "twoSidedLighting": instance.GetTwoSidedLighting(),
                "lightFollowCamera": instance.GetLightFollowCamera(),
                "layer": instance.GetLayer(),
                "preserveColorBuffer": instance.GetPreserveColorBuffer(),
                "preserveDepthBuffer": instance.GetPreserveDepthBuffer(),
                "nearClippingPlaneTolerance": instance.GetNearClippingPlaneTolerance(),
                "clippingRangeExpansion": instance.GetClippingRangeExpansion(),
                "useShadows": instance.GetUseShadows(),
                "useDepthPeeling": instance.GetUseDepthPeeling(),
                "occlusionRatio": instance.GetOcclusionRatio(),
                "maximumNumberOfPeels": instance.GetMaximumNumberOfPeels(),
                "interactive": instance.GetInteractive(),
            },
            "dependencies": dependencies,
            "calls": calls,
        }

    return None


# -----------------------------------------------------------------------------


def cameraSerializer(parent, instance, objId, context, depth):
    return {
        "parent": getReferenceId(parent),
        "id": objId,
        "type": class_name(instance),
        "properties": {
            "focalPoint": instance.GetFocalPoint(),
            "position": instance.GetPosition(),
            "viewUp": instance.GetViewUp(),
            "clippingRange": instance.GetClippingRange(),
        },
    }


# -----------------------------------------------------------------------------


def lightTypeToString(value):
    """
    #define VTK_LIGHT_TYPE_HEADLIGHT    1
    #define VTK_LIGHT_TYPE_CAMERA_LIGHT 2
    #define VTK_LIGHT_TYPE_SCENE_LIGHT  3

    'HeadLight';
    'SceneLight';
    'CameraLight'
    """
    if value == 1:
        return "HeadLight"
    elif value == 2:
        return "CameraLight"

    return "SceneLight"


def lightSerializer(parent, instance, objId, context, depth):
    return {
        "parent": getReferenceId(parent),
        "id": objId,
        "type": class_name(instance),
        "properties": {
            # 'specularColor': instance.GetSpecularColor(),
            # 'ambientColor': instance.GetAmbientColor(),
            "switch": instance.GetSwitch(),
            "intensity": instance.GetIntensity(),
            "color": instance.GetDiffuseColor(),
            "position": instance.GetPosition(),
            "focalPoint": instance.GetFocalPoint(),
            "positional": instance.GetPositional(),
            "exponent": instance.GetExponent(),
            "coneAngle": instance.GetConeAngle(),
            "attenuationValues": instance.GetAttenuationValues(),
            "lightType": lightTypeToString(instance.GetLightType()),
            "shadowAttenuation": instance.GetShadowAttenuation(),
        },
    }


# -----------------------------------------------------------------------------


def renderWindowSerializer(parent, instance, objId, context, depth):
    dependencies = []
    rendererIds = []

    rendererCollection = instance.GetRenderers()
    for rIdx in range(rendererCollection.GetNumberOfItems()):
        # Grab the next vtkRenderer
        renderer = rendererCollection.GetItemAsObject(rIdx)
        rendererId = getReferenceId(renderer)
        rendererInstance = serializeInstance(
            instance, renderer, rendererId, context, depth + 1
        )
        if rendererInstance:
            dependencies.append(rendererInstance)
            rendererIds.append(rendererId)

    calls = context.buildDependencyCallList(
        objId, rendererIds, "addRenderer", "removeRenderer"
    )

    return {
        "parent": getReferenceId(parent),
        "id": objId,
        "type": class_name(instance),
        "properties": {"numberOfLayers": instance.GetNumberOfLayers()},
        "dependencies": dependencies,
        "calls": calls,
        "mtime": instance.GetMTime(),
    }



from math import *

# -----------------------------------------------------------------------------
# Set of helper functions
# -----------------------------------------------------------------------------


def normalize(vect, tolerance=0.00001):
    mag2 = sum(n * n for n in vect)
    if abs(mag2 - 1.0) > tolerance:
        mag = sqrt(mag2)
        vect = tuple(n / mag for n in vect)
    return vect


def q_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return w, x, y, z


def q_conjugate(q):
    w, x, y, z = q
    return (w, -x, -y, -z)


def qv_mult(q1, v1):
    q2 = (0.0,) + v1
    return q_mult(q_mult(q1, q2), q_conjugate(q1))[1:]


def axisangle_to_q(v, theta):
    v = normalize(v)
    x, y, z = v
    theta /= 2
    w = cos(theta)
    x = x * sin(theta)
    y = y * sin(theta)
    z = z * sin(theta)
    return w, x, y, z


def vectProduct(axisA, axisB):
    xa, ya, za = axisA
    xb, yb, zb = axisB
    normalVect = (ya * zb - za * yb, za * xb - xa * zb, xa * yb - ya * xb)
    normalVect = normalize(normalVect)
    return normalVect


def dotProduct(vecA, vecB):
    return (vecA[0] * vecB[0]) + (vecA[1] * vecB[1]) + (vecA[2] * vecB[2])


def rotate(axis, angle, center, point):
    angleInRad = 3.141592654 * angle / 180.0
    rotation = axisangle_to_q(axis, angleInRad)
    tPoint = tuple((point[i] - center[i]) for i in range(3))
    rtPoint = qv_mult(rotation, tPoint)
    rPoint = tuple((rtPoint[i] + center[i]) for i in range(3))
    return rPoint


# -----------------------------------------------------------------------------
# Spherical Camera
# -----------------------------------------------------------------------------


class SphericalCamera(object):
    def __init__(
        self, dataHandler, focalPoint, position, phiAxis, phiAngles, thetaAngles
    ):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.thetaBind = {
            "mouse": {
                "drag": {"modifier": 0, "coordinate": 1, "step": 30, "orientation": 1}
            }
        }
        self.phiBind = {
            "mouse": {
                "drag": {"modifier": 0, "coordinate": 0, "step": 30, "orientation": 1}
            }
        }

        # Convert to serializable type
        fp = tuple(i for i in focalPoint)

        # Register arguments to the data handler
        if len(phiAngles) > 1 and phiAngles[-1] + phiAngles[1] == 360:
            self.dataHandler.registerArgument(
                priority=0,
                name="phi",
                values=phiAngles,
                ui="slider",
                loop="modulo",
                bind=self.phiBind,
            )
        else:
            self.dataHandler.registerArgument(
                priority=0, name="phi", values=phiAngles, ui="slider", bind=self.phiBind
            )
        if thetaAngles[0] < 0 and thetaAngles[0] >= -90:
            idx = 0
            for theta in thetaAngles:
                if theta < 0:
                    idx += 1

            self.dataHandler.registerArgument(
                priority=0,
                name="theta",
                values=[(x + 90) for x in thetaAngles],
                ui="slider",
                default=idx,
                bind=self.thetaBind,
            )
        else:
            self.dataHandler.registerArgument(
                priority=0,
                name="theta",
                values=thetaAngles,
                ui="slider",
                bind=self.thetaBind,
            )

        # Compute all camera settings
        for theta in thetaAngles:
            for phi in phiAngles:
                phiPos = rotate(phiAxis, -phi, fp, position)
                thetaAxis = vectProduct(
                    phiAxis, tuple(fp[i] - phiPos[i] for i in range(3))
                )
                thetaPhiPos = rotate(thetaAxis, theta, fp, phiPos)
                viewUp = rotate(thetaAxis, theta, (0, 0, 0), phiAxis)

                self.cameraSettings.append(
                    {
                        "theta": theta,
                        "thetaIdx": thetaAngles.index(theta),
                        "phi": phi,
                        "phiIdx": phiAngles.index(phi),
                        "focalPoint": fp,
                        "position": thetaPhiPos,
                        "viewUp": viewUp,
                    }
                )

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ["theta", "phi"]
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for cameraData in self.cameraSettings:
            self.dataHandler.setArguments(
                phi=cameraData["phiIdx"], theta=cameraData["thetaIdx"]
            )
            yield cameraData


# -----------------------------------------------------------------------------
# Cylindrical Camera
# -----------------------------------------------------------------------------


class CylindricalCamera(object):
    def __init__(
        self,
        dataHandler,
        focalPoint,
        position,
        rotationAxis,
        phiAngles,
        translationValues,
    ):
        self.dataHandler = dataHandler
        self.cameraSettings = []

        # Register arguments to the data handler
        self.dataHandler.registerArgument(
            priority=0, name="phi", values=phiAngles, ui="slider", loop="modulo"
        )
        self.dataHandler.registerArgument(
            priority=0, name="n_pos", values=translationValues, ui="slider"
        )

        # Compute all camera settings
        for translation in translationValues:
            for phi in phiAngles:
                phiPos = rotate(rotationAxis, phi, focalPoint, position)
                newfocalPoint = tuple(
                    focalPoint[i] + (translation * rotationAxis[i]) for i in range(3)
                )
                transPhiPoint = tuple(
                    phiPos[i] + (translation * rotationAxis[i]) for i in range(3)
                )

                self.cameraSettings.append(
                    {
                        "n_pos": translation,
                        "n_posIdx": translationValues.index(translation),
                        "phi": phi,
                        "phiIdx": phiAngles.index(phi),
                        "focalPoint": newfocalPoint,
                        "position": transPhiPoint,
                        "viewUp": rotationAxis,
                    }
                )

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ["n_pos", "phi"]
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for cameraData in self.cameraSettings:
            self.dataHandler.setArguments(
                phi=cameraData["phiIdx"], n_pos=cameraData["n_posIdx"]
            )
            yield cameraData


# -----------------------------------------------------------------------------
# MultiView Cube Camera
# -----------------------------------------------------------------------------


class CubeCamera(object):

    # positions = [ { position: [x,y,z], args: { i: 1, j: 0, k: 7 } }, ... ]
    def __init__(self, dataHandler, viewForward, viewUp, positions):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.viewForward = viewForward
        self.viewUp = viewUp
        self.rightDirection = vectProduct(viewForward, viewUp)
        self.positions = positions

        # Register arguments to the data handler
        self.dataHandler.registerArgument(
            priority=0, name="orientation", values=["f", "b", "r", "l", "u", "d"]
        )

        # Register arguments to id position
        self.args = {}
        for pos in positions:
            for key in pos["args"]:
                if key not in self.args:
                    self.args[key] = {}
                self.args[key][pos["args"][key]] = True

        for key in self.args:
            self.args[key] = sorted(self.args[key], key=lambda k: int(k))

        self.keyList = self.args.keys()
        for key in self.args:
            self.dataHandler.registerArgument(
                priority=1, name=key, values=self.args[key]
            )

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ["orientation"]
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for pos in self.positions:
            cameraData = {
                "position": pos["position"],
            }

            print("=" * 80)
            for key in pos["args"]:
                idx = self.args[key].index(pos["args"][key])
                self.dataHandler.setArguments(**{key: idx})
                print(key, idx)

            print("position", cameraData["position"])

            # front
            cameraData["focalPoint"] = [
                (cameraData["position"][i] + self.viewForward[i]) for i in range(3)
            ]
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["orientation"] = "front"
            self.dataHandler.setArguments(orientation=0)
            yield cameraData

            # back
            cameraData["focalPoint"] = [
                (cameraData["position"][i] - self.viewForward[i]) for i in range(3)
            ]
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["orientation"] = "back"
            self.dataHandler.setArguments(orientation=1)
            yield cameraData

            # right
            self.dataHandler.setArguments(orientation=2)
            cameraData["focalPoint"] = [
                (cameraData["position"][i] + self.rightDirection[i]) for i in range(3)
            ]
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["orientation"] = "right"
            yield cameraData

            # left
            self.dataHandler.setArguments(orientation=3)
            cameraData["focalPoint"] = [
                (cameraData["position"][i] - self.rightDirection[i]) for i in range(3)
            ]
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["orientation"] = "left"
            yield cameraData

            # up
            self.dataHandler.setArguments(orientation=4)
            cameraData["focalPoint"] = [
                (cameraData["position"][i] + self.viewUp[i]) for i in range(3)
            ]
            cameraData["viewUp"] = [(-self.viewForward[i]) for i in range(3)]
            cameraData["orientation"] = "up"
            yield cameraData

            # doww
            self.dataHandler.setArguments(orientation=5)
            cameraData["focalPoint"] = [
                (cameraData["position"][i] - self.viewUp[i]) for i in range(3)
            ]
            cameraData["viewUp"] = [self.viewForward[i] for i in range(3)]
            cameraData["orientation"] = "down"
            yield cameraData


# -----------------------------------------------------------------------------
# MultiView Cube Camera
# -----------------------------------------------------------------------------


class StereoCubeCamera(object):

    # positions = [ { position: [x,y,z], args: { i: 1, j: 0, k: 7 } }, ... ]
    def __init__(self, dataHandler, viewForward, viewUp, positions, eyeSpacing):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.viewForward = viewForward
        self.viewUp = viewUp
        self.rightDirection = vectProduct(viewForward, viewUp)
        self.positions = positions
        self.eyeSpacing = eyeSpacing

        # Register arguments to the data handler
        self.dataHandler.registerArgument(
            priority=0, name="orientation", values=["f", "b", "r", "l", "u", "d"]
        )
        self.dataHandler.registerArgument(
            priority=0, name="eye", values=["left", "right"]
        )

        # Register arguments to id position
        self.args = {}
        for pos in positions:
            for key in pos["args"]:
                if key not in self.args:
                    self.args[key] = {}
                self.args[key][pos["args"][key]] = True

        for key in self.args:
            self.args[key] = sorted(self.args[key], key=lambda k: int(k))

        self.keyList = self.args.keys()
        for key in self.args:
            self.dataHandler.registerArgument(
                priority=1, name=key, values=self.args[key]
            )

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ["orientation"]
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for pos in self.positions:
            cameraData = {}

            for key in pos["args"]:
                idx = self.args[key].index(pos["args"][key])
                self.dataHandler.setArguments(**{key: idx})

            # front
            cameraData["orientation"] = "front"
            self.dataHandler.setArguments(orientation=0)
            deltaVect = [
                (v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection
            ]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] - deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] + self.viewForward[i] - deltaVect[i])
                for i in range(3)
            ]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] + deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] + self.viewForward[i] + deltaVect[i])
                for i in range(3)
            ]
            yield cameraData

            # back
            cameraData["orientation"] = "back"
            self.dataHandler.setArguments(orientation=1)
            deltaVect = [
                -(v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection
            ]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] - deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] - self.viewForward[i] - deltaVect[i])
                for i in range(3)
            ]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] + deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] - self.viewForward[i] + deltaVect[i])
                for i in range(3)
            ]
            yield cameraData

            # right
            self.dataHandler.setArguments(orientation=2)
            cameraData["orientation"] = "right"
            deltaVect = [-(v * float(self.eyeSpacing) * 0.5) for v in self.viewForward]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] - deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] + self.rightDirection[i] - deltaVect[i])
                for i in range(3)
            ]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] + deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] + self.rightDirection[i] + deltaVect[i])
                for i in range(3)
            ]
            yield cameraData

            # left
            self.dataHandler.setArguments(orientation=3)
            cameraData["orientation"] = "left"
            deltaVect = [(v * float(self.eyeSpacing) * 0.5) for v in self.viewForward]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] - deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] - self.rightDirection[i] - deltaVect[i])
                for i in range(3)
            ]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData["viewUp"] = [self.viewUp[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] + deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] - self.rightDirection[i] + deltaVect[i])
                for i in range(3)
            ]
            yield cameraData

            # up
            self.dataHandler.setArguments(orientation=4)
            cameraData["orientation"] = "up"
            deltaVect = [
                (v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection
            ]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData["viewUp"] = [(-self.viewForward[i]) for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] - deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] + self.viewUp[i] - deltaVect[i]) for i in range(3)
            ]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData["viewUp"] = [(-self.viewForward[i]) for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] + deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] + self.viewUp[i] + deltaVect[i]) for i in range(3)
            ]
            yield cameraData

            # doww
            self.dataHandler.setArguments(orientation=5)
            cameraData["orientation"] = "down"
            deltaVect = [
                (v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection
            ]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData["viewUp"] = [self.viewForward[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] - deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] - self.viewUp[i] - deltaVect[i]) for i in range(3)
            ]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData["viewUp"] = [self.viewForward[i] for i in range(3)]
            cameraData["position"] = [
                (pos["position"][idx] + deltaVect[idx]) for idx in range(3)
            ]
            cameraData["focalPoint"] = [
                (pos["position"][i] - self.viewUp[i] + deltaVect[i]) for i in range(3)
            ]
            yield cameraData


# -----------------------------------------------------------------------------
# MultiView Camera
# -----------------------------------------------------------------------------


class MultiViewCamera(object):
    def __init__(self, dataHandler):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.positionNames = []

    def registerViewPoint(self, name, focalPoint, position, viewUp):
        self.cameraSettings.append(
            {
                "name": name,
                "nameIdx": len(self.positionNames),
                "focalPoint": focalPoint,
                "position": position,
                "viewUp": viewUp,
            }
        )
        self.positionNames.append(name)
        self.dataHandler.registerArgument(
            priority=0, name="multiView", values=self.positionNames
        )
        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ["multiView"]
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for cameraData in self.cameraSettings:
            self.dataHandler.setArguments(multiView=cameraData["nameIdx"])
            yield cameraData


# -----------------------------------------------------------------------------
# Helper methods
# -----------------------------------------------------------------------------


def update_camera(renderer, cameraData):
    camera = renderer.GetActiveCamera()
    camera.SetPosition(cameraData["position"])
    camera.SetFocalPoint(cameraData["focalPoint"])
    camera.SetViewUp(cameraData["viewUp"])


def create_spherical_camera(renderer, dataHandler, phiValues, thetaValues):
    camera = renderer.GetActiveCamera()
    return SphericalCamera(
        dataHandler,
        camera.GetFocalPoint(),
        camera.GetPosition(),
        camera.GetViewUp(),
        phiValues,
        thetaValues,
    )


def create_cylindrical_camera(renderer, dataHandler, phiValues, translationValues):
    camera = renderer.GetActiveCamera()
    return CylindricalCamera(
        dataHandler,
        camera.GetFocalPoint(),
        camera.GetPosition(),
        camera.GetViewUp(),
        phiValues,
        translationValues,
    )



import base64
import json
import re
import os
import shutil
import sys
import zipfile

try:
    import zlib

    compression = zipfile.ZIP_DEFLATED
except:
    compression = zipfile.ZIP_STORED

# -----------------------------------------------------------------------------


def convertDirectoryToZipFile(directoryPath):
    if os.path.isfile(directoryPath):
        return

    zipFilePath = "%s.zip" % directoryPath
    zf = zipfile.ZipFile(zipFilePath, mode="w")

    try:
        for dirName, subdirList, fileList in os.walk(directoryPath):
            for fname in fileList:
                fullPath = os.path.join(dirName, fname)
                relPath = "%s" % (os.path.relpath(fullPath, directoryPath))
                zf.write(fullPath, arcname=relPath, compress_type=compression)
    finally:
        zf.close()

    shutil.rmtree(directoryPath)
    shutil.move(zipFilePath, directoryPath)


# -----------------------------------------------------------------------------


def addDataToViewer(dataPath, srcHtmlPath, disableGirder=False):
    if os.path.isfile(dataPath) and os.path.exists(srcHtmlPath):
        dstDir = os.path.dirname(dataPath)
        dstHtmlPath = os.path.join(dstDir, "%s.html" % os.path.basename(dataPath)[:-6])

        # Extract data as base64
        with open(dataPath, "rb") as data:
            dataContent = data.read()
            base64Content = base64.b64encode(dataContent)
            base64Content = base64Content.decode().replace("\n", "")

        # Create new output file
        with open(srcHtmlPath, mode="r", encoding="utf-8") as srcHtml:
            with open(dstHtmlPath, mode="w", encoding="utf-8") as dstHtml:
                for line in srcHtml:
                    if disableGirder and "</title>" in line:
                        dstHtml.write(
                            """
    <script>
        // Force reloading the page if we want to disable girder before anything else.
        const urlParams = new URLSearchParams(window.location.search);
        if (urlParams.get('noGirder') != 'true') {
            urlParams.set('noGirder', 'true');
            window.location.search = urlParams;
        }
    </script>
                                      """
                        )
                    if "</body>" in line:
                        dstHtml.write("<script>\n")
                        dstHtml.write('var contentToLoad = "%s";\n\n' % base64Content)
                        dstHtml.write(
                            'Glance.importBase64Dataset("%s" , contentToLoad, glanceInstance.proxyManager);\n'
                            % os.path.basename(dataPath)
                        )
                        dstHtml.write("glanceInstance.showApp();\n")
                        dstHtml.write("</script>\n")

                    dstHtml.write(line)


# -----------------------------------------------------------------------------


def numericSorted(l):
    """Numerically sort a list of strings."""

    # pattern to split name into numeric and non-numeric parts
    splitter_pattern = re.compile('([0-9]+|[^0-9]+)')

    def keyfunc(name):
        """Sorting key for numeric sorting."""
        split_name = re.findall(splitter_pattern, name)
        # one-liner to convert numeric parts into integers
        split_name = list(map(lambda x: int(x) if x.isdigit() else x, split_name))
        # ensure that list begins with a string to avoid string<->int compare
        if split_name and isinstance(split_name[0], int):
            split_name.insert(0, '')
        return split_name

    # return the numerically sorted list
    return sorted(l, key=keyfunc)


# -----------------------------------------------------------------------------


def zipAllTimeSteps(directoryPath):
    if os.path.isfile(directoryPath):
        return

    class UrlCounterDict(dict):
        Counter = 0

        def GetUrlName(self, name):
            if name not in self.keys():
                self[name] = str(objNameToUrls.Counter)
                self.Counter = self.Counter + 1
            return self[name]

    def InitIndex(sourcePath, destObj):
        with open(sourcePath, "r") as sourceFile:
            sourceData = sourceFile.read()
            sourceObj = json.loads(sourceData)
            for key in sourceObj:
                destObj[key] = sourceObj[key]
            # remove vtkHttpDataSetReader information
            for obj in destObj["scene"]:
                obj.pop(obj["type"])
                obj.pop("type")

    def getUrlToNameDictionary(indexObj):
        urls = {}
        for obj in indexObj["scene"]:
            urls[obj[obj["type"]]["url"]] = obj["name"]
        return urls

    def addDirectoryToZip(
        dirname, zipobj, storedData, rootIdx, timeStep, objNameToUrls
    ):
        # Update root index.json file from index.json of this timestep
        with open(os.path.join(dirname, "index.json"), "r") as currentIdxFile:
            currentIdx = json.loads(currentIdxFile.read())
            urlToName = getUrlToNameDictionary(currentIdx)
            rootTimeStepSection = rootIdx["animation"]["timeSteps"][timeStep]
            for key in currentIdx:
                if key == "scene" or key == "version":
                    continue
                rootTimeStepSection[key] = currentIdx[key]
            for obj in currentIdx["scene"]:
                objName = obj["name"]
                rootTimeStepSection[objName] = {}
                rootTimeStepSection[objName]["actor"] = obj["actor"]
                rootTimeStepSection[objName]["actorRotation"] = obj["actorRotation"]
                rootTimeStepSection[objName]["mapper"] = obj["mapper"]
                rootTimeStepSection[objName]["property"] = obj["property"]

        # For every object in the current timestep
        for folder in sorted(os.listdir(dirname)):
            currentItem = os.path.join(dirname, folder)
            if os.path.isdir(currentItem) is False:
                continue
            # Write all data array of the current timestep in the archive
            for filename in os.listdir(os.path.join(currentItem, "data")):
                fullpath = os.path.join(currentItem, "data", filename)
                if os.path.isfile(fullpath) and filename not in storedData:
                    storedData.add(filename)
                    relPath = os.path.join("data", filename)
                    zipobj.write(fullpath, arcname=relPath, compress_type=compression)
            # Write the index.json containing pointers to these data arrays
            # while replacing every basepath as '../../data'
            objIndexFilePath = os.path.join(dirname, folder, "index.json")
            with open(objIndexFilePath, "r") as objIndexFile:
                objIndexObjData = json.loads(objIndexFile.read())
            for elm in objIndexObjData.keys():
                try:
                    if "ref" in objIndexObjData[elm].keys():
                        objIndexObjData[elm]["ref"]["basepath"] = "../../data"
                    if "arrays" in objIndexObjData[elm].keys():
                        for array in objIndexObjData[elm]["arrays"]:
                            array["data"]["ref"]["basepath"] = "../../data"
                except AttributeError:
                    continue
            currentObjName = urlToName[folder]
            objIndexRelPath = os.path.join(
                objNameToUrls.GetUrlName(currentObjName), str(timeStep), "index.json"
            )
            zipobj.writestr(
                objIndexRelPath,
                json.dumps(objIndexObjData, indent=2),
                compress_type=compression,
            )

    # ---

    zipFilePath = "%s.zip" % directoryPath
    currentDirectory = os.path.abspath(os.path.join(directoryPath, os.pardir))
    rootIndexPath = os.path.join(currentDirectory, "index.json")
    rootIndexFile = open(rootIndexPath, "r")
    rootIndexObj = json.loads(rootIndexFile.read())

    zf = zipfile.ZipFile(zipFilePath, mode="w")
    try:
        # We copy the scene from an index of a specific timestep to the root index
        # Scenes should all have the same objects so only do it for the first one
        isSceneInitialized = False
        # currentlyAddedData set stores hashes of every data we already added to the
        # vtkjs archive to prevent data duplication
        currentlyAddedData = set()
        # Regex that folders storing timestep data from paraview should follow
        reg = re.compile(r"^" + os.path.basename(directoryPath) + r"\.[0-9]+$")
        # We assume an object will not be deleted from a timestep to another so we create a generic index.json for each object
        genericIndexObj = {}
        genericIndexObj["series"] = []
        timeStep = 0
        for item in rootIndexObj["animation"]["timeSteps"]:
            genericIndexObj["series"].append({})
            genericIndexObj["series"][timeStep]["url"] = str(timeStep)
            genericIndexObj["series"][timeStep]["timeStep"] = float(item["time"])
            timeStep = timeStep + 1
        # Keep track of the url for every object
        objNameToUrls = UrlCounterDict()

        timeStep = 0
        # zip all timestep directories
        for folder in numericSorted(os.listdir(currentDirectory)):
            fullPath = os.path.join(currentDirectory, folder)
            if os.path.isdir(fullPath) and reg.match(folder):
                if not isSceneInitialized:
                    InitIndex(os.path.join(fullPath, "index.json"), rootIndexObj)
                    isSceneInitialized = True
                addDirectoryToZip(
                    fullPath,
                    zf,
                    currentlyAddedData,
                    rootIndexObj,
                    timeStep,
                    objNameToUrls,
                )
                shutil.rmtree(fullPath)
                timeStep = timeStep + 1

        # Write every index.json holding time information for each object
        for name in objNameToUrls:
            zf.writestr(
                os.path.join(objNameToUrls[name], "index.json"),
                json.dumps(genericIndexObj, indent=2),
                compress_type=compression,
            )

        # Update root index.json urls and write it in the archive
        for obj in rootIndexObj["scene"]:
            obj["id"] = obj["name"]
            obj["type"] = "vtkHttpDataSetSeriesReader"
            obj["vtkHttpDataSetSeriesReader"] = {}
            obj["vtkHttpDataSetSeriesReader"]["url"] = objNameToUrls[obj["name"]]
        zf.writestr(
            "index.json", json.dumps(rootIndexObj, indent=2), compress_type=compression
        )

        rootIndexFile.close()
        os.remove(rootIndexPath)

    finally:
        zf.close()

    shutil.move(zipFilePath, directoryPath)


# -----------------------------------------------------------------------------
# Main
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(
            "Usage: directoryToFile /path/to/directory.vtkjs [/path/to/ParaViewGlance.html]"
        )
    else:
        fileName = sys.argv[1]
        convertDirectoryToZipFile(fileName)

        if len(sys.argv) == 3:
            addDataToViewer(fileName, sys.argv[2])
